/*setup includes*/
#include <windows.h>
#include <stdio.h>
#include <commctrl.h>
#include "tasks.h"
#include "../ls-b24/lsapi/lsapi.h"

/*setup constants*/
const char AppName[] = "Tasks";
const int MAXTASKS = 16;
const int MAXICONS = 64;

/*setup variables*/
IconList iconTag[MAXICONS];
IgnoreList ignoreTag[MAXTASKS];
PixList pixTag[MAXICONS];
TasksSettings ts;
TaskType tasks[MAXTASKS];

BOOL lbtnPressed, movingTask, taskWrap, bgImgNone;
char path[MAXLEN];
HCURSOR normalCursor, movingCursor;
HINSTANCE dll;
HWND parent, mainWin, deskTop, hints, curTask;
int numIcon, numIgnore, numPix, numTasks, prevTasks;
int screenHeight, screenWidth;
POINT oldCursorPos;
UINT timer;

/*declare all the functions*/
BOOL GetTasksIgnore(char *ignoreTitle);
BOOL IsAppWindow(HWND hWNd);
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
HICON GetIconFromWindow(HWND hWnD, BOOL bBigIcon);
int GetTaskByWnd(HWND Hwnd);
int GetUnMovedNrTasks(void);
void CreateHints(HWND hWnd, char *txt, RECT *r);
void CreateTask(int i, HWND handle);
void DrawBorders(COLORREF lightColor, COLORREF darkColor, COLORREF bgColor, RECT *r);
void DeleteTask(int i);
void GatherTasks(void);
void GetNewTaskPos(int i);
void GetTasks(void);
void GetTasksIcon(int n);
void GetTasksPix(int n);
void PaintTask(int i);
void ReadConfig(void);
void ReadTasksIcon(void);
void ReadTasksIgnore(void);
void ReadTasksPix(void);
void ReArrangeTasks(BOOL all);
void RemoveHints(HWND hWnd);
void TaskPosition(int i, int tmpNumTasks);
void ToggleTasks(void);
void HideTasks(void);
void ShowTasks(void);

/*and here's all the functions*/
int initModule(HWND ParentWnd, HINSTANCE dllInst, wharfDataType* wd)
{
  return initModuleEx (ParentWnd, dllInst, wd->lsPath);
}

TasksSettings ReadSettings()
{
  TasksSettings settings;
  char tmp[MAXLEN], tmp2[MAXLEN], tmp3[MAXLEN]; //, tmp4[MAXLEN], tmp5[MAXLEN];

  GetRCString("TasksDirection", tmp, "right", MAXLEN);
	  settings.Direction = 2;  /*only needed if user enters an invalid parameter*/
	if (strcmp(tmp, "up") == 0)
	  settings.Direction = 1;
	else if (strcmp(tmp, "right") == 0)
	  settings.Direction = 2;
	else if (strcmp(tmp, "down") == 0)
	  settings.Direction = 3;
	else if (strcmp(tmp, "left") == 0)
	  settings.Direction = 4;
//	tmp[sizeof(tmp)] = 0;
  settings.PaddingX = GetRCInt("TasksPaddingX", 0);
  settings.PaddingY = GetRCInt("TasksPaddingY", 0);
  settings.WrapCount = GetRCInt("TasksWrapCount", MAXTASKS + 1);
	if (settings.WrapCount > MAXTASKS || settings.WrapCount < 2)
	{
	  settings.WrapCount = MAXTASKS;
	  taskWrap = false;
	}
	else
	  taskWrap = true;
  GetRCString("TasksWrapDirection", tmp2, "down", MAXLEN);
	  settings.WrapDirection = 3;  /*only needed if user enters an invalid parameter*/
	if (strcmp(tmp2, "up") == 0)
	  settings.WrapDirection = 1;
	else if (strcmp(tmp2, "right") == 0)
	  settings.WrapDirection = 2;
	else if (strcmp(tmp2, "down") == 0)
	  settings.WrapDirection = 3;
	else if (strcmp(tmp2, "left") == 0)
	  settings.WrapDirection = 4;
//	tmp[sizeof(tmp)] = 0;

  settings.Height = GetRCInt("TasksHeight", 32);
    if (settings.Height < 0) settings.Height = 32;
  settings.Width = GetRCInt("TasksWidth", 32);
    if (settings.Width < 0) settings.Width = 32;
  settings.X = GetRCInt("TasksX", 0);
  settings.Y = GetRCInt("TasksY", 0);

  settings.MinimizedX = GetRCInt("TasksMinimizedX", 0);
  settings.MinimizedY = GetRCInt("TasksMinimizedY", 0);

  settings.IconSize = GetRCInt("TasksIconSize", 16);
    if (settings.IconSize < 0) settings.IconSize = 16;
  settings.IconX = GetRCInt("TasksIconX", 9);
  settings.IconY = GetRCInt("TasksIconY", 9);

  settings.PixX = GetRCInt("TasksPixX", 0);
  settings.PixY = GetRCInt("TasksPixY", 0);
		
  settings.AllwaysOnTop = GetRCBool("TasksAllwaysOnTop", TRUE);
  settings.AutoArrange = GetRCBool("TasksAutoArrange", TRUE);
  settings.GroupMinimized = GetRCBool("TasksGroupMinimized", TRUE);
  settings.OnlyMinimized = GetRCBool("TasksOnlyMinimized", TRUE);
  settings.MoveAll = GetRCBool("TasksMoveAll", TRUE);
  settings.StartHidden = GetRCBool("TasksStartHidden", TRUE);

  settings.NoHints = GetRCBool("TasksNoHints", TRUE);
  settings.NoMove = GetRCBool("TasksNoMove", TRUE);
  settings.HighLightSelected = GetRCBool("TasksHighLightSelected", TRUE);
  settings.ShadeMinimized = GetRCBool("TasksShadeMinimized", TRUE);

  settings.BgColor = GetRCColor("TasksBgColor", 0xFFFF00);
  settings.DarkColor = GetRCColor("TasksDarkColor", 0xFF00FF);
  settings.LightColor = GetRCColor("TasksLightColor", 0xFF00FF);
  settings.SelDarkColor = GetRCColor("TasksSelDarkColor", 0x000000);
  settings.SelLightColor = GetRCColor("TasksSelLightColor", 0xFFFFFF);
  settings.MinShadeColor = GetRCColor("TasksMinShadeColor", 0x0000FF);

/*Insert TasksDefault here!!!*/
  GetRCString("TasksBackImage", tmp3, ".none", MAXLEN);
	if (strcmp(".none", tmp3))
	{
	  settings.BackImage = LoadLSImage(tmp3, NULL);
	  if (strcmp(".none", tmp3) == 0)
	    bgImgNone = true;
	  else
	    bgImgNone = false;
	}
	else
	{
	  settings.BackImage = NULL;
	  bgImgNone = false;
	}
//	strcpy (tmp3, NULL);
//	tmp[sizeof(tmp)] = 0;


  settings.TitleIconX = GetRCInt("TasksTitleIconX", 1);
  settings.TitleIconY = GetRCInt("TasksTitleIconY", 1);
  settings.TitlePixX = GetRCInt("TasksTitlePixX", 1);
  settings.TitlePixY = GetRCInt("TasksTitlePixY", 1);

  settings.TitleHeight = GetRCInt("TasksTitleHeight", 14);
    if (settings.TitleHeight < 0) settings.TitleHeight = 14;
  settings.TitleWidth = GetRCInt("TasksTitleWidth", 30);
    if (settings.TitleWidth < 0) settings.TitleWidth = 30;
  settings.TitleX = GetRCInt("TasksTitleX", 1);
  settings.TitleY = GetRCInt("TasksTitleY", 18);

  settings.TitleFontSize = GetRCInt("TasksTitleFontSize", 10);
    if (settings.TitleFontSize < 2) settings.TitleFontSize = 10;

  settings.TitleAlways = GetRCBool("TasksTitleAlways", TRUE);
  settings.TitleOnlyMinimized = GetRCBool("TasksTitleOnlyMinimized", TRUE);

  GetRCString("TasksTitleFont", settings.TitleFont, "Arial", MAXLEN);

  settings.TitleBgColor = GetRCColor("TasksTitleBgColor", 0xFF00FF);
  settings.TitleDarkColor = GetRCColor("TasksTitleDarkColor", 0xFF00FF);
  settings.TitleFontColor = GetRCColor("TasksTitleFontColor", 0xFFFFFF);
  settings.TitleLightColor = GetRCColor("TasksTitleLightColor", 0xFF00FF);
  return settings;
}

int initModuleEx(HWND ParentWnd, HINSTANCE dllInst, LPCSTR szPath)
{
  int i;
  RECT r;
  dll = dllInst;
  strcpy(path, szPath);
  parent = ParentWnd;
  GetClientRect(GetDesktopWindow(),&r);
  screenWidth = r.right;
  screenHeight = r.bottom;
  movingTask = false;
  deskTop = NULL;
  deskTop = FindWindow("DesktopBackgroundClass", NULL);
  if (!deskTop)
    deskTop = GetDesktopWindow();
  {
	WNDCLASS wc;
    memset(&wc,0,sizeof(wc));
	wc.lpfnWndProc = WndProc;
	wc.hInstance = dllInst;
	wc.lpszClassName = AppName;
	wc.style = CS_DBLCLKS;
	if (!RegisterClass(&wc)) 
	{
	  MessageBox(parent,"Error registering window class","Tasks",MB_OK);
	  return 1;
	}
  }
  normalCursor = GetCursor();
  if (!normalCursor)
	LoadCursor(NULL, IDC_ARROW);
  movingCursor = LoadCursor(NULL, IDC_SIZEALL);
  for (i = 0; i < MAXTASKS; i++)
  {
	tasks[i].hMenu = NULL;
	tasks[i].hwnd = NULL;
	tasks[i].pix = NULL;
	tasks[i].appHandle = NULL;
	tasks[i].appTitle = NULL;
	tasks[i].icon = NULL;
  }
  ts = ReadSettings();
  ReadTasksIcon();
  ReadTasksIgnore();
  ReadTasksPix();
  mainWin = CreateWindowEx(WS_EX_TOOLWINDOW, AppName, "", WS_CHILD, 0, 0, 1, 1, deskTop, NULL, dll,	NULL);
  if (!mainWin) 
  {						   
    MessageBox(parent,"Error creating window","Tasks",MB_OK);
    return 1;
  }
  if (!ts.NoHints)
  {
	hints = CreateWindow(TOOLTIPS_CLASS,(LPSTR) NULL,TTS_ALWAYSTIP,	CW_USEDEFAULT,	CW_USEDEFAULT,	CW_USEDEFAULT,CW_USEDEFAULT,NULL,(HMENU) NULL, dll,	NULL);
	if (!hints) 
	{
	  MessageBox(parent, "Error creating hints window", "Tasks", MB_OK);
	  return 1;
	}
	SetWindowPos(hints, HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE|SWP_NOACTIVATE);
  }
  SetWindowLong(mainWin, GWL_USERDATA, magicDWord);
  ShowWindow(mainWin, SW_HIDE);
  AddBangCommand("!GATHERTASKS", GatherTasks);
  AddBangCommand("!TOGGLETASKS", ToggleTasks);
  AddBangCommand("!HIDETASKS", HideTasks);
  AddBangCommand("!SHOWTASKS", ShowTasks);
  timer = SetTimer(mainWin, 0, 250, NULL);
  return 0;
}

void quitModule(HINSTANCE dllInst)
{
  int i;
  if (timer)
    KillTimer(NULL, timer);
  for(i = 0; i < numTasks; i++)
    DeleteTask(i);
  for(i = 0; i < numPix; i++)
  {
	delete [] pixTag[numPix].match;
	delete [] pixTag[numPix].pix;
  }
  for(i = 0; i < numIcon; i++)
  {
	delete [] iconTag[numIcon].match;
	delete [] iconTag[numIcon].name;
  }
  for(i = 0; i < numIgnore; i++)
	delete [] ignoreTag[numIgnore].match;
  if (ts.BackImage)
    DeleteObject(ts.BackImage);
  DestroyWindow(mainWin);
  if (!ts.NoHints) 
	DestroyWindow (hints);
  UnregisterClass(AppName,dllInst);
  return;
}

BOOL IsAppWindow(HWND hWNd)
{
	BOOL bResult = TRUE;
	LONG nStyle;
	HWND hOwner;
	HWND hParent;

	/*if it is a WS_CHILD or not WS_VISIBLE, fail it*/
	nStyle = GetWindowLong(hWNd, GWL_STYLE);	
	if((nStyle & WS_CHILD) || !(nStyle & WS_VISIBLE))
		bResult = FALSE;

	/*if the window is a WS_EX_TOOLWINDOW fail it*/
	nStyle = GetWindowLong(hWNd, GWL_EXSTYLE);
	if(nStyle & WS_EX_TOOLWINDOW)
		bResult = FALSE;

	/*If this has an owner, then only accept this window
	  if the partent is not accepted*/
	hOwner = GetWindow(hWNd, GW_OWNER);
	if(hOwner && bResult)
		bResult = !IsAppWindow(hOwner);

	/*if it has a parent, fail*/
	hParent = GetParent(hWNd);
	if(hParent)
		bResult = FALSE;

	if(GetWindowLong(hWNd, GWL_USERDATA) == 0x49474541)
		bResult = FALSE;

	return bResult;
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  switch (message)
  {
	case WM_TIMER:
	{
	  curTask = GetForegroundWindow();
	  GetTasks();
      return 0;
	}
    case WM_QUERYENDSESSION:
	{
      return SendMessage(parent,message,wParam,lParam);
	}
	case WM_CREATE:
	  return 0;
	case WM_ERASEBKGND:
	  return 1; 
	case WM_PAINT:
	{
      int n = GetTaskByWnd(hwnd);
      if (n != -1)
		PaintTask(n);
	  return 0;
	}
	case WM_LBUTTONDOWN:
	{
	  int n = GetTaskByWnd(hwnd);
	  if (n != -1)
	  {
	    GetCursorPos(&oldCursorPos);
	    SetCapture(tasks[n].hwnd);
		lbtnPressed = true;
		movingTask = false;
	  }
	  return 0;
	}
	case WM_LBUTTONUP:
	{
	  int n = GetTaskByWnd(hwnd);
	  if (n != -1)
	  {
  	    POINT cursorPos;
	    GetCursorPos(&cursorPos);
  	    ReleaseCapture();
		lbtnPressed = false;
		if (GetCursor() != normalCursor)
	      SetCursor(normalCursor);
		if (abs(oldCursorPos.x - cursorPos.x) <= 5 && abs(oldCursorPos.y - cursorPos.y) <= 5 && tasks[n].hwnd == WindowFromPoint(cursorPos)) // click or canceled move
		{
		  if (!movingTask)
		  {
			if (IsIconic(tasks[n].appHandle))
			  ShowWindow(tasks[n].appHandle, SW_RESTORE);
			else if (tasks[n].appHandle == curTask)
			  ShowWindow(tasks[n].appHandle, SW_MINIMIZE);
			if (!IsIconic(tasks[n].appHandle))
			{
			  SendMessage(parent, LM_BRINGTOFRONT, 0, (long)tasks[n].appHandle);
			  SetForegroundWindow(tasks[n].appHandle);
			}
		  }
		  else /*canceled a move, force a redraw (I hate this, shouldnt be necesary)*/
		  {
			ShowWindow(tasks[n].hwnd, SW_HIDE);
		    ShowWindow(tasks[n].hwnd, SW_SHOWNORMAL);
		  }
		}
		else if (!ts.NoMove) /*move*/
		{
		  tasks[n].x -= oldCursorPos.x - cursorPos.x;
		  tasks[n].y -= oldCursorPos.y - cursorPos.y;
		  tasks[n].IsMoved = true;
		  SetWindowPos(tasks[n].hwnd, 0, tasks[n].x, tasks[n].y, ts.Width, ts.Height, 0); // SWP_NOZORDER|SWP_NOSIZE
		  ShowWindow(tasks[n].hwnd, SW_HIDE); /*stupid redraw again*/
		  ShowWindow(tasks[n].hwnd, SW_SHOWNORMAL);
		  if (ts.AutoArrange)
			ReArrangeTasks(false);
		}
	  }
	  return 0;
	}
	case WM_MOUSEMOVE:
	{
	  int n = GetTaskByWnd(hwnd);
	  if (lbtnPressed && !ts.NoMove)
	  {
        POINT cursorPos;
		if (n != -1)
		{
	      GetCursorPos(&cursorPos);
	      if (abs(oldCursorPos.x - cursorPos.x) > 5 || abs(oldCursorPos.y - cursorPos.y) > 5 ||  tasks[n].hwnd != WindowFromPoint(cursorPos))
		  {
		    if (GetCursor() != movingCursor)
		      SetCursor(movingCursor);
			SetWindowPos(tasks[n].hwnd, 0, tasks[n].x - (oldCursorPos.x - cursorPos.x), tasks[n].y - (oldCursorPos.y - cursorPos.y), ts.Width, ts.Height, 0); //SWP_NOZORDER|SWP_NOSIZE
			movingTask = true;
		  }
		  else
		  {
		    if (GetCursor() != normalCursor)
		      SetCursor(normalCursor);
			SetWindowPos(tasks[n].hwnd, 0, tasks[n].x, tasks[n].y, ts.Width, ts.Height, 0); // SWP_NOZORDER|SWP_NOSIZE
		  }
		}
	  }
	  else
	  {
        if (!ts.NoHints) 
		{
		  MSG TooltipMessage;
		  TooltipMessage.hwnd=tasks[n].hwnd;
		  TooltipMessage.message=message;
		  TooltipMessage.wParam=wParam;
		  TooltipMessage.lParam=lParam;
		  SendMessage(hints,TTM_RELAYEVENT,0,(LPARAM)&TooltipMessage);
		  SetWindowPos(hints, HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE|SWP_NOACTIVATE);
		}
	  }
	  return 0;
	}
	case WM_MOUSEACTIVATE:
	{
		PostMessage(parent, 9185, 0, 0); /*Tells litestep it has been selected*/
		SendMessage(parent, 9183, (int)HIWORD(lParam), (int)LOWORD(lParam)); /*Close any popup open*/
		if (HIWORD(lParam) == WM_LBUTTONDOWN) /*te 10/30/98, Windows seems to be eating this?*/
			PostMessage(hwnd, WM_LBUTTONDOWN, 0, 0);
		else if (HIWORD(lParam) == WM_RBUTTONDOWN) /*te 10/30/98, Windows seems to be eating this?*/
			PostMessage(hwnd, WM_RBUTTONDOWN, 0, 0);
		return MA_ACTIVATE;
	}
	case WM_ACTIVATE:
	{
		SendMessage(parent, 9185, 0, 0); /*Tells litestep it has been selected*/
		if (LOWORD(wParam)) SetActiveWindow(parent);
		return 0;
	}
	case WM_RBUTTONUP: /*Open system popup menu - Thanks to Fahims lsTasks src! :)*/
	{
		int n, x, y;
		POINT pt;
		BOOL result;

		GetCursorPos(&pt);
		x = pt.x;
		y = pt.y;
		n = GetTaskByWnd(hwnd);
		if (n != -1)
		{
//			tasks[n].hMenu = GetSystemMenu(tasks[n].appHandle, FALSE);  //not needed anymore (done in CreateTasks)
//			SendMessage(tasks[n].appHandle, 0x313, 0, GetMessagePos());  // testing this, works only on certain apps. uncomment below for normal usage
			result = TrackPopupMenu(tasks[n].hMenu, TPM_RETURNCMD|TPM_RIGHTBUTTON, x, y, 0, hwnd, NULL);
			if (result)
			{
				if (result != SC_MINIMIZE && result != SC_ICON)
					SetForegroundWindow(tasks[n].appHandle);
				SendMessage(tasks[n].appHandle, WM_SYSCOMMAND, (WPARAM)result, MAKELPARAM(x,y));
			}
//			DestroyMenu(hMenu); // not needed anymore (Done in DeleteTasks)
		}
		return 0;
	}
	case WM_RBUTTONDOWN: /*Close popup menu*/
		return 0;
	case WM_MBUTTONDOWN: /*LiteStep poup menu*/
		SendMessage(parent, 9182, (int)HIWORD(lParam), (int)LOWORD(lParam));
		return 0;
	case WM_KEYDOWN: 
	case WM_KEYUP:
	{	
	  PostMessage(parent,message,wParam,lParam);
	}
	return 0;
  }
  return DefWindowProc(hwnd,message,wParam,lParam);
}

HICON GetIconFromWindow(HWND hWnD, BOOL bBigIcon) /*taken from the re5ource popups, much credit to re5ource :)*/
{
	HICON a;

	if (!bBigIcon) /* thanks again to fahim's lstasks src*/
    {
	    SendMessageTimeout(hWnD, WM_GETICON, 0, 0, 0, 1000, (unsigned long*)&a);
		if (!a) a = (HICON)GetClassLong(hWnD, GCL_HICONSM);
		if (!a) SendMessageTimeout(hWnD, WM_GETICON, 1, 0, 0, 1000, (unsigned long*)&a);
		if (!a) a = (HICON)GetClassLong(hWnD, GCL_HICON);
		if (!a) SendMessageTimeout(hWnD, WM_QUERYDRAGICON, 0, 0, 0, 1000, (unsigned long*)&a);
    } else {
	    SendMessageTimeout(hWnD, WM_GETICON, 1, 0, 0, 1000, (unsigned long*)&a);
		if (!a) a = (HICON)GetClassLong(hWnD, GCL_HICON);
	    if (!a) SendMessageTimeout(hWnD, WM_QUERYDRAGICON, 0, 0, 0, 1000, (unsigned long*)&a);
		if (!a) SendMessageTimeout(hWnD, WM_GETICON, 0, 0, 0, 1000, (unsigned long*)&a);
		if (!a) a = (HICON)GetClassLong(hWnD, GCL_HICONSM);
    }
	if (!a) a = LoadIcon(NULL, MAKEINTRESOURCE(IDI_APPLICATION));
	return a;

}
//	a = ExtractIcon (dll, "c:\\windows\\explorer.exe",3);
/*   HICON hIcon = NULL;

   int nBigOrSmall = bBigIcon ? ICON_BIG : ICON_SMALL;
   int nBigOrSmall2 = bBigIcon ? GCL_HICON : GCL_HICONSM;

   SendMessageTimeout(hWnD, WM_GETICON, nBigOrSmall, 0, SMTO_ABORTIFHUNG, 1000, (PULONG)&hIcon);
   if(!hIcon) hIcon = (HICON)GetClassLong(hWnD, nBigOrSmall2);
   if(!hIcon) SendMessageTimeout(hWnD, WM_GETICON, nBigOrSmall, 1, SMTO_ABORTIFHUNG, 1000, (PULONG)&hIcon);
   if(!hIcon) hIcon = (HICON)GetClassLong(hWnD, nBigOrSmall2);
   if(!hIcon) SendMessageTimeout(hWnD, WM_QUERYDRAGICON, nBigOrSmall, 0, SMTO_ABORTIFHUNG, 1000, (PULONG)&hIcon);
   
   return hIcon;
   */

int GetTaskByWnd(HWND Hwnd)
{
  int i;
  for (i = 0;i < numTasks ; i++)
    if (tasks[i].hwnd == Hwnd)
	  return i;
  return -1;
}

int GetUnMovedNrTasks(void)
{
  int result = 0;
  for (int i = 0; i < numTasks; i++)
    if (!tasks[i].IsMoved)
	  result++;
  return result;
}

void GetTasksIcon(int n)
{
  int i;
  for (i = 0; i < numIcon; i++)
  {
	if (strstr(tasks[n].appTitle, iconTag[i].match))
	{
	  tasks[n].icon = ExtractIcon(dll,iconTag[i].name,iconTag[i].icon);
	  if (!tasks[n].icon) tasks[n].icon = LoadIcon(NULL, MAKEINTRESOURCE(IDI_APPLICATION));
	  return;
	}
  }
  tasks[n].icon = GetIconFromWindow(tasks[n].appHandle, (ts.IconSize > 23) ? true: false);
  return;
}

BOOL GetTasksIgnore(char *ignoreTitle)
{
  int i;
  for (i = 0; i < numIgnore; i++)
  {
	if (strstr(ignoreTitle, ignoreTag[i].match))
	  return true;
  }
  return false;
}

void GetTasksPix(int n)
{
  int i;
  for (i = 0; i < numPix; i++)
  {
	if (strstr(tasks[n].appTitle, pixTag[i].match))
	{
	  tasks[n].pix = LoadLSImage(pixTag[i].pix, NULL);
	  if (tasks[n].pix) tasks[n].icon = NULL;
	  return;
	}
  }
  return;
}

void CreateHints(HWND hWnd, char *txt, RECT *r)
{ /*from the original ls code (shortcuts)*/
  TOOLINFO ti;    /*tool information*/
  if (ts.NoHints)
	return;
  ti.cbSize = sizeof(TOOLINFO);
  ti.uFlags = TTF_SUBCLASS;
  ti.hwnd = hWnd;
  ti.hinst = dll;
  ti.uId = 0;
  ti.lpszText = txt;
  ti.rect = *r;
  SendMessage(hints, TTM_ADDTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
  return;
}

void CreateTask(int i, HWND handle)
{
  RECT r;
  HWND hWnD = NULL;
  char title[MAXLEN];
  hWnD = CreateWindowEx(WS_EX_TOOLWINDOW, AppName, "", WS_CHILD, 0, 0,	ts.Width, ts.Height, deskTop, NULL, dll, NULL);
  if (!hWnD) 
  {						   
    MessageBox(parent,"Error creating window","Tasks",MB_OK);
    return;
  }
  tasks[i].hwnd = hWnD;
  tasks[i].appHandle = handle;
  SetWindowLong(tasks[i].hwnd, GWL_USERDATA, magicDWord);
  GetWindowText(handle, (char *)&title, MAXLEN);
  tasks[i].appTitle = new char[strlen(title) + 1];
  strcpy(tasks[i].appTitle, title);
  tasks[i].Active = (GetForegroundWindow() == handle);
//  tasks[i].Ignore = GetTasksIgnore(i);
  tasks[i].IsIconic = IsIconic(handle);
  tasks[i].IsMoved  = false;
  GetNewTaskPos(i);
  GetTasksIcon(i);
  GetTasksPix(i);
  tasks[i].hMenu = GetSystemMenu(tasks[i].appHandle, FALSE);
  SetWindowPos(tasks[i].hwnd, 0, tasks[i].x, tasks[i].y, ts.Width, ts.Height, 0);
  if (ts.StartHidden)
	ShowWindow(tasks[i].hwnd, SW_HIDE);
  else
	ShowWindow(tasks[i].hwnd, SW_SHOWNORMAL);
  GetClientRect(tasks[i].hwnd, &r);
  CreateHints(tasks[i].hwnd, tasks[i].appTitle, &r);
  InvalidateRect(tasks[i].hwnd, &r, TRUE);
  return;
}

void DeleteTask(int i)
{
  RemoveHints(tasks[i].hwnd);
  if (tasks[i].hwnd)
    DestroyWindow(tasks[i].hwnd);
  if (tasks[i].pix)
    DeleteObject(tasks[i].pix);
  if (tasks[i].appTitle)
    delete [] tasks[i].appTitle;
  if (tasks[i].hMenu)
	DestroyMenu(tasks[i].hMenu);
  tasks[i].appHandle = NULL;
  tasks[i].hMenu = NULL;
  tasks[i].hwnd = NULL;
  tasks[i].pix = NULL;
  tasks[i].icon = NULL;  
  tasks[i].appTitle = NULL;
  return;
}

void GatherTasks(void)
{
  for (int i = 0; i < numTasks; i++)
	tasks[i].IsMoved = false;
  ReArrangeTasks(true);
  return;
}

void GetNewTaskPos(int i)
{
  TaskPosition(i, (ts.AutoArrange ? GetUnMovedNrTasks(): numTasks));
  return;
}

void TaskPosition(int i, int tmpNumTasks)
{
  int wrapLine = 0, resetLine = 0, checkLine = 0;
  wrapLine = tmpNumTasks / ts.WrapCount;
  checkLine = tmpNumTasks - ts.WrapCount * wrapLine;
  if (checkLine == 1)
	resetLine = 1;
  else
	resetLine = tmpNumTasks - ts.WrapCount * wrapLine;
  if (ts.Direction == 1) /*up*/
  {
	if (taskWrap)
	{
	  if (ts.Y >= 0)
	    tasks[i].y = ts.Y - ts.Height - resetLine * (ts.Height + ts.PaddingY);
	  else /*now accepts negative values*/
	    tasks[i].y = screenHeight + ts.Y - ts.Height - resetLine * (ts.Height + ts.PaddingY);
	  if (ts.WrapDirection == 4) /*left*/
	  {
		if (ts.X >= 0)
		  tasks[i].x = ts.X - ts.Width - wrapLine * (ts.Width + ts.PaddingX);
		else /*now accepts negative values*/
		  tasks[i].x = screenWidth + ts.X - ts.Width - wrapLine * (ts.Width + ts.PaddingX);
	  }
	  else /*defaults to right, even if user inputs invalid up or down setting*/
	  {
	    if (ts.X >= 0)
	      tasks[i].x = ts.X + wrapLine * (ts.Width + ts.PaddingX);
	    else /*now accepts negative values*/
	      tasks[i].x = screenWidth + ts.X + wrapLine * (ts.Width + ts.PaddingX);
	  }
	}
	else
	{
	  if (ts.X >= 0)
	    tasks[i].x = ts.X;
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + ts.X;
	  if (ts.Y >= 0)
	    tasks[i].y = ts.Y - ts.Height - tmpNumTasks * (ts.Height + ts.PaddingY);
	  else /*now accepts negative values*/
	    tasks[i].y = screenHeight + ts.Y - ts.Height - tmpNumTasks * (ts.Height + ts.PaddingY);
	}
  }
  else if (ts.Direction == 2) /*right*/
  {
    if (taskWrap)
	{
	  if (ts.X >= 0)
		tasks[i].x = ts.X + resetLine * (ts.Width + ts.PaddingX);
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + ts.X + resetLine * (ts.Width + ts.PaddingX);
	  if (ts.WrapDirection == 1) /*up*/
	  {
	    if (ts.Y >= 0)
	      tasks[i].y = ts.Y - ts.Height - wrapLine * (ts.Height + ts.PaddingY);
	    else /*now accepts negative values*/
		  tasks[i].y = screenHeight + ts.Y - ts.Height - wrapLine * (ts.Height + ts.PaddingY);
	  }
	  else /*defaults to down, even if user inputs invalid left or right setting*/
	  {
	    if (ts.Y >= 0)
	      tasks[i].y = ts.Y + wrapLine * (ts.Height + ts.PaddingY);
	    else /*now accepts negative values*/
	      tasks[i].y = screenHeight + ts.Y + wrapLine * (ts.Height + ts.PaddingY);
	  }
	}
	else
	{
	  if (ts.X >= 0)
		tasks[i].x = ts.X + tmpNumTasks * (ts.Width + ts.PaddingX);
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + ts.X + tmpNumTasks * (ts.Width + ts.PaddingX);
	  if (ts.Y >= 0)
		tasks[i].y = ts.Y;
	  else /*now accepts negative values*/
		tasks[i].y = screenHeight + ts.Y;
	}
  }
  else if (ts.Direction == 3) /*down*/
  {
    if (taskWrap)
	{
	  if (ts.Y >= 0)
	    tasks[i].y = ts.Y + resetLine * (ts.Height + ts.PaddingY);
	  else /*now accepts negative values*/
	    tasks[i].y = screenHeight + ts.Y + resetLine * (ts.Height + ts.PaddingY);
	  if (ts.WrapDirection == 4) /*left*/
	  {
		if (ts.X >= 0)
		  tasks[i].x = ts.X - ts.Width - wrapLine * (ts.Width + ts.PaddingX);
		else /*now accepts negative values*/
		  tasks[i].x = screenWidth + ts.X - ts.Width - wrapLine * (ts.Width + ts.PaddingX);
	  }
	  else /*defaults to right, even if user inputs invalid up or down setting*/
	  {
	    if (ts.X >= 0)
	      tasks[i].x = ts.X + wrapLine * (ts.Width + ts.PaddingX);
	    else /*now accepts negative values*/
	      tasks[i].x = screenWidth + ts.X + wrapLine * (ts.Width + ts.PaddingX);
	  }
	}
	else
	{
	  if (ts.X >= 0)
	    tasks[i].x = ts.X;
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + ts.X;
	  if (ts.Y >= 0)
	    tasks[i].y = ts.Y + tmpNumTasks * (ts.Height + ts.PaddingY);
	  else /*now accepts negative values*/
	    tasks[i].y = screenHeight + ts.Y + tmpNumTasks * (ts.Height + ts.PaddingY);
	}
  }
  else if (ts.Direction == 4) /*left*/
  {
	if (taskWrap)
	{
	  if (ts.X >= 0)
	    tasks[i].x = ts.X - ts.Width - resetLine * (ts.Width + ts.PaddingX);
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + ts.X - ts.Width - resetLine * (ts.Width + ts.PaddingX);
	  if (ts.WrapDirection == 1) /*up*/
	  {
	    if (ts.Y >= 0)
	      tasks[i].y = ts.Y - ts.Height - wrapLine * (ts.Height + ts.PaddingY);
	    else /*now accepts negative values*/
		  tasks[i].y = screenHeight + ts.Y - ts.Height - wrapLine * (ts.Height + ts.PaddingY);
	  }
	  else /*defaults to down, even if user inputs invalid left or right setting*/
	  {
	    if (ts.Y >= 0)
	      tasks[i].y = ts.Y + wrapLine * (ts.Height + ts.PaddingY);
	    else /*now accepts negative values*/
	      tasks[i].y = screenHeight + ts.Y + wrapLine * (ts.Height + ts.PaddingY);
	  }
	}
	else
	{
	  if (ts.X >= 0)
	    tasks[i].x = ts.X - ts.Width - tmpNumTasks * (ts.Width + ts.PaddingX);
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + ts.X - ts.Width - tmpNumTasks * (ts.Width + ts.PaddingX);
      if (ts.Y >= 0)
        tasks[i].y = ts.Y;
      else /*now accepts negative values*/
        tasks[i].y = screenHeight + ts.Y;
	}
  }
  return;
}

void GetTasks(void)
{
  windowType * winList = NULL;
  char tmpTitle[MAXLEN];
  int m, n, maxWin;
  prevTasks = numTasks;
  numTasks = 0;		  
  winList = (windowType *)SendMessage(parent, LM_WINLIST, 0, 0);
  maxWin = (int) SendMessage(parent, LM_WINLIST, 1, 0);
  for (n = 0; n < maxWin && winList[n].Handle; n++)
  {
	GetWindowText(winList[n].Handle, (char *)&tmpTitle, MAXLEN);
	if ((winList[n].Handle != NULL && numTasks < MAXTASKS) && (!ts.OnlyMinimized || IsIconic(winList[n].Handle)) && IsAppWindow(winList[n].Handle) && !GetTasksIgnore(tmpTitle))
	{
	  if (winList[n].Handle != tasks[numTasks].appHandle)
	  { 
		DeleteTask(numTasks);
	    CreateTask(numTasks, winList[n].Handle);
	    numTasks++;
	  }
	  else if (tasks[numTasks].IsIconic != IsIconic(winList[n].Handle))
	  {
		tasks[numTasks].IsIconic = IsIconic(winList[n].Handle);
		if (!ts.StartHidden)
		{
		  ShowWindow(tasks[numTasks].hwnd, SW_HIDE);
		  ShowWindow(tasks[numTasks].hwnd, SW_SHOWNORMAL);		
		  /*there must be a better way to force a redraw*/
		}
		numTasks++;
	  }
	  else if (strcmp(tmpTitle, tasks[numTasks].appTitle) != 0)
	  {
		  if (!ts.NoHints)
		  {
			TOOLINFO ti;
		    ti.cbSize = sizeof(TOOLINFO);
	  	    ti.hwnd = tasks[n].hwnd;
	        ti.uId = 0;
	        ti.lpszText = tmpTitle;
		    SendMessage(hints, TTM_UPDATETIPTEXT, 0, (LPARAM) (LPTOOLINFO) &ti);
		    delete [] tasks[numTasks].appTitle;
		    tasks[numTasks].appTitle = new char[strlen(tmpTitle) + 1];
		    strcpy(tasks[numTasks].appTitle, tmpTitle);
		    numTasks++;
		  }
	  }
	  else if (tasks[numTasks].Active != (curTask == tasks[numTasks].appHandle))
	  {
		if (curTask == tasks[numTasks].appHandle)
		  tasks[numTasks].Active = true;
		else
		  tasks[numTasks].Active = false;
		/*Darn lame way to Refresh again! bah*/
		ShowWindow(tasks[numTasks].hwnd, SW_HIDE);
		ShowWindow(tasks[numTasks].hwnd, SW_SHOWNORMAL);
		numTasks++;
		//MessageBox(parent,"painting","Tasks",MB_OK);
	  }
	  else
	    numTasks++;
	}
  }
  for (m = numTasks; m < prevTasks; m++)
	DeleteTask(m);
  return;
}

void PaintTask(int i)
{
  int posX, posY;
  PAINTSTRUCT ps;
  BITMAP bmInfo;
  HDC hdc = BeginPaint(tasks[i].hwnd,&ps);
  HDC src = CreateCompatibleDC(NULL);
  if (ts.BackImage)
  {
    SelectObject(src, ts.BackImage);
    TransparentBltLS(hdc, 0, 0, ts.Width, ts.Height, src, 0, 0, RGB(255, 0, 255));
  }
  if((!tasks[i].icon && !tasks[i].pix) || (!ts.BackImage && !bgImgNone))
  {
	RECT r;
	GetClientRect(tasks[i].hwnd, &r);
	if (ts.BgColor != 0x00FF00FF)
	{
	  HBRUSH backBrush;
	  backBrush = CreateSolidBrush(ts.BgColor);
	  FillRect(hdc, &r, backBrush);
	  DeleteObject(backBrush);
	}
  }
  if (tasks[i].pix)
  {
    GetObject(tasks[i].pix, sizeof(bmInfo), &bmInfo);
/*    if (bmInfo.bmWidth > (ts.Width - ts.PixX))
      bmInfo.bmWidth = ts.Width - ts.PixX;
    if (bmInfo.bmHeight > (ts.Height - ts.PixY)) 
      bmInfo.bmHeight = ts.Height - ts.PixY;*/
	posX = ts.PixX;
	posY = ts.PixY;
	if (ts.TitleAlways || (ts.TitleOnlyMinimized && tasks[i].IsIconic))
	{
	  posX = ts.TitlePixX;
	  posY = ts.TitlePixY;
	}
    SelectObject(src, tasks[i].pix);
	TransparentBltLS(hdc, posX, posY, bmInfo.bmWidth, bmInfo.bmHeight, src, 0, 0, RGB(255,0,255));
  }
  if (tasks[i].icon)
  {
	posX = ts.IconX;
	posY = ts.IconY;
	if (ts.TitleAlways || (ts.TitleOnlyMinimized && tasks[i].IsIconic))
	{
	  posX = ts.TitleIconX;
	  posY = ts.TitleIconY;
	}
	DrawIconEx(hdc, posX, posY, tasks[i].icon, ts.IconSize, ts.IconSize, 0, NULL, DI_NORMAL);
  }
  if (ts.TitleAlways || (ts.TitleOnlyMinimized && tasks[i].IsIconic))
  {
	RECT r;
	GetClientRect(tasks[i].hwnd, &r);
	  r.top = ts.TitleY;
	  r.left = ts.TitleX;
	  r.bottom = ts.TitleHeight + ts.TitleY;
	  r.right = ts.TitleWidth + ts.TitleX;
	if (ts.TitleBgColor != 0x00FF00FF)
	{
	  HBRUSH backBrush;
	  backBrush = CreateSolidBrush(ts.TitleBgColor);
	  FillRect(hdc, &r, backBrush);
	  DeleteObject(backBrush);
	}
	if ((ts.TitleLightColor != 0x00FF00FF) && (ts.TitleDarkColor != 0x00FF00FF))
	{
	  HPEN pen, oldpen;
	  /*Light Color Bevel (left/top)*/
	  pen = CreatePen(PS_SOLID, 1, ts.TitleLightColor);
	  oldpen = (HPEN) SelectObject(hdc, pen);
	  MoveToEx(hdc, r.left, r.bottom - 1, NULL);
	  LineTo(hdc, r.left, r.top);
	  LineTo(hdc, r.right - 1, r.top);
	  SelectObject(hdc, oldpen);
	  DeleteObject(pen);
	  /*Dark Color Bevel (bottom/right)*/
	  pen = CreatePen(PS_SOLID, 1, ts.TitleDarkColor);
	  oldpen = (HPEN) SelectObject(hdc, pen);
	  MoveToEx(hdc, r.right - 1, r.top, NULL);
	  LineTo(hdc, r.right - 1, r.bottom - 1);
	  LineTo(hdc, r.left, r.bottom - 1);
	  SelectObject(hdc, oldpen);
	  DeleteObject(pen);
	}
    char title[MAXLEN];
    HFONT titleFont, oldFont;
    int len = strlen(tasks[i].appTitle);
	r.top += 1;
	r.bottom -= 1;
	r.left += 2;
	r.right -= 2;
    titleFont = CreateFont(ts.TitleFontSize, 0, 0, 0, 0, FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, ts.TitleFont);
    oldFont = (HFONT) SelectObject(hdc, titleFont);
    SetBkMode(hdc, TRANSPARENT);
    SetTextColor(hdc, ts.TitleFontColor);
    GetWindowText(tasks[i].appHandle, (char *)&title, MAXLEN);
    DrawText(hdc, title, len, &r, DT_VCENTER|DT_SINGLELINE|DT_END_ELLIPSIS);
    SelectObject(hdc, oldFont);
    DeleteObject(titleFont);
  }
  if ((ts.LightColor != 0x00FF00FF) && (ts.DarkColor != 0x00FF00FF) && (!ts.HighLightSelected || !tasks[i].Active))
  {
	RECT r;
	GetClientRect(tasks[i].hwnd, &r);
	HPEN pen, oldpen;
	/*Light Color Bevel (left/top)*/
	pen = CreatePen(PS_SOLID, 1, ts.LightColor);
	oldpen = (HPEN) SelectObject(hdc, pen);
	MoveToEx(hdc, r.left, r.bottom - 1, NULL);
	LineTo(hdc, r.left, r.top);
	LineTo(hdc, r.right - 1, r.top);
	SelectObject(hdc, oldpen);
	DeleteObject(pen);
	/*Dark Color Bevel (bottom/right)*/
	pen = CreatePen(PS_SOLID, 1, ts.DarkColor);
	oldpen = (HPEN) SelectObject(hdc, pen);
	MoveToEx(hdc, r.right - 1, r.top, NULL);
	LineTo(hdc, r.right - 1, r.bottom - 1);
	LineTo(hdc, r.left, r.bottom - 1);
	SelectObject(hdc, oldpen);
	DeleteObject(pen);
  }
  else if ((ts.SelLightColor != 0x00FF00FF) && (ts.SelDarkColor != 0x00FF00FF) && ts.HighLightSelected && tasks[i].Active)
  {
	RECT r;
	GetClientRect(tasks[i].hwnd, &r);
	HPEN pen, oldpen;
	/*Light Color Bevel (left/top)*/
	pen = CreatePen(PS_SOLID, 1, ts.SelLightColor);
	oldpen = (HPEN) SelectObject(hdc, pen);
	MoveToEx(hdc, r.left, r.bottom - 1, NULL);
	LineTo(hdc, r.left, r.top);
	LineTo(hdc, r.right - 1, r.top);
	SelectObject(hdc, oldpen);
	DeleteObject(pen);
	/*Dark Color Bevel (bottom/right)*/
	pen = CreatePen(PS_SOLID, 1, ts.SelDarkColor);
	oldpen = (HPEN) SelectObject(hdc, pen);
	MoveToEx(hdc, r.right - 1, r.top, NULL);
	LineTo(hdc, r.right - 1, r.bottom - 1);
	LineTo(hdc, r.left, r.bottom - 1);
	SelectObject(hdc, oldpen);
	DeleteObject(pen);
  }
  DeleteDC(src);
  EndPaint(tasks[i].hwnd,&ps);
  return;
}

void ReadTasksIcon(void)
{
  FILE * f;
  unsigned int i;
  char * token[4];
  char buffer[1024];

  numIcon = 0;
  f = LCOpen(NULL);
  if (!f) 
	return;
  for (i = 0; i < 4; i++)
    token[i] = new char[MAXLEN];
  while (LCReadNextConfig (f, "*TasksIcon", buffer, sizeof (buffer)))
  {
	 int count;
	 count = LCTokenize (buffer, token, 4, NULL);
	 if (count >= 4)
	 {
	   iconTag[numIcon].match = new char[strlen(token[1])+1];
	   iconTag[numIcon].name = new char[strlen(token[2])+1];
	   strcpy(iconTag[numIcon].match, token[1]);
	   strcpy(iconTag[numIcon].name, token[2]);
	   iconTag[numIcon].icon = atoi(token[3]);
	   numIcon++;
	 }
  }
  LCClose(f);
  for (i = 0; i < 4; i++)
    delete [] token[i];
  return;
}

void ReadTasksIgnore(void)
{
  FILE * f;
  unsigned int i;
  char * token[2];
  char buffer[1024];

  numIgnore = 0;
  f = LCOpen(NULL);
  if (!f) 
	return;
  for (i = 0; i < 2; i++)
    token[i] = new char[MAXLEN];
  while (LCReadNextConfig (f, "*TasksIgnore", buffer, sizeof (buffer)))
  {
	 int count;
	 count = LCTokenize (buffer, token, 2, NULL);
	 if (count >= 2)
	 {
	   ignoreTag[numIgnore].match = new char[strlen(token[1])+1];
	   strcpy(ignoreTag[numIgnore].match, token[1]);
	   numIgnore++;
	 }
  }
  LCClose(f);
  for (i = 0; i < 2; i++)
    delete [] token[i];
  return;
}

void ReadTasksPix(void)
{
  FILE * f;
  unsigned int i;
  char * token[3];
  char buffer[1024];

  numPix = 0;
  f = LCOpen(NULL);
  if (!f) 
	return;
  for (i = 0; i < 3; i++)
    token[i] = new char[MAXLEN];
  while (LCReadNextConfig (f, "*TasksPix", buffer, sizeof (buffer)))
  {
	 int count;
	 count = LCTokenize (buffer, token, 3, NULL);
	 if (count >= 3)
	 {
	   pixTag[numPix].match = new char[strlen(token[1])+1];
	   pixTag[numPix].pix = new char[strlen(token[2])+1];
	   strcpy(pixTag[numPix].match, token[1]);
	   strcpy(pixTag[numPix].pix, token[2]);
	   numPix++;
	 }
  }
  LCClose(f);
  for (i = 0; i < 3; i++)
    delete [] token[i];
  return;
}

void ReArrangeTasks(BOOL all)
{
  int j = 0;
  for (int i = 0; i < numTasks; i++)
  {
	if (all || !tasks[i].IsMoved)
	{
	  TaskPosition(i, j);
      SetWindowPos(tasks[i].hwnd, 0, tasks[i].x, tasks[i].y, ts.Width, ts.Height, 0);
	  j++;
	}
  }
  all = false;
  return;
}

void RemoveHints(HWND hWnd)
{ /*from the original ls code (shortcuts)*/
	TOOLINFO ti;    /*tool information*/
	RECT r={0,0,0,0};

	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = 0;
	ti.hwnd = hWnd;
	ti.hinst = dll;
	ti.uId = 0;
	ti.lpszText = NULL;
	ti.rect = r;

	SendMessage(hints, TTM_DELTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
  return;
}

void ToggleTasks(void)
{
  for (int i = 0; i < numTasks; i++)
  {
	if (ts.StartHidden)
	  ShowWindow(tasks[i].hwnd, SW_SHOWNORMAL);
	else
	  ShowWindow(tasks[i].hwnd, SW_HIDE);
  }
  ts.StartHidden = !ts.StartHidden;
  return;
}

void HideTasks(void)
{
  for (int i = 0; i < numTasks; i++)
  {
	ShowWindow(tasks[i].hwnd, SW_HIDE);
  }
  ts.StartHidden = true;
  return;
}

void ShowTasks(void)
{
  for (int i = 0; i < numTasks; i++)
  {
	ShowWindow(tasks[i].hwnd, SW_SHOWNORMAL);
  }
  ts.StartHidden = false;
  return;
}